@@ -30,7 +30,12 @@
 #undef LOCAL_DEBUG
 
 static char *link_tmpdir = NULL;
+#ifdef G_OS_WIN32
+static LinkNetIdType use_local_host = LINK_NET_ID_IS_IPADDR;
+#else
 static LinkNetIdType use_local_host = LINK_NET_ID_IS_FQDN;
+#endif
+static const char *fixed_host_net_id = NULL;
 
 /*
  * make_local_tmpdir:
@@ -58,7 +63,7 @@ make_local_tmpdir (const char *dirname)
 				g_error ("Can not stat %s\n", dirname);
 
 #if !defined (__CYGWIN__) && !defined(_WIN32)
-			if (statbuf.st_uid != getuid ())
+			if (getuid() != 0 && statbuf.st_uid != getuid ())
 				g_error ("Owner of %s is not the current user\n", dirname);
 
 			if ((statbuf.st_mode & (S_IRWXG|S_IRWXO)) ||
@@ -134,44 +139,71 @@ link_get_tmpdir (void)
 
 #if defined(AF_INET) || defined(AF_INET6) || defined (AF_UNIX)
 
+#ifndef G_OS_WIN32
+static void
+get_first_non_local_ipaddr(char *buf, 
+			   size_t len)
+{
+	int sock;
+	const char *retv;
+	struct if_nameindex *ifname_idx_array;
+	struct if_nameindex *ifname_idx;
+	struct sockaddr_in sin;
+	struct ifreq ifr;
+
+	*buf = '\0';
+
+	sock = socket(AF_INET, SOCK_DGRAM, 0);
+	if (-1 == sock)
+		return;
+
+	ifname_idx_array = if_nameindex();
+	for (ifname_idx = ifname_idx_array; ifname_idx && ifname_idx->if_name && ifname_idx->if_name[0]; ifname_idx++) {
+		strncpy(ifr.ifr_name, ifname_idx->if_name, IFNAMSIZ);
+		if (ioctl(sock, SIOCGIFADDR, &ifr)) 
+			continue;
+
+		memcpy(&sin, &(ifr.ifr_addr), sizeof(struct sockaddr_in));
+		retv = (const char*)inet_ntoa(sin.sin_addr);
+		retv = strcmp("127.0.0.1", retv) ? retv : NULL;
+		if (retv) {
+			strncpy(buf, (const char*)inet_ntoa(sin.sin_addr), len);
+			break;
+		}
+	}
+
+	if (ifname_idx_array)
+		if_freenameindex(ifname_idx_array);
+
+	if (-1 != sock)
+		close(sock);
+}
+#endif
+
 static char *
 get_netid(LinkNetIdType which, 
 	  char *buf, 
 	  size_t len)
 {
-	if (LINK_NET_ID_IS_LOCAL == which)
-		return strncpy(buf, "localhost", len);
-
-	if (LINK_NET_ID_IS_IPADDR == which) {
+	if (LINK_NET_ID_IS_LOCAL == which) {
 #ifndef G_OS_WIN32
-		struct sockaddr_in *adr = NULL;
-		struct ifreq my_ifreqs[2];
-		struct ifconf my_ifconf;
-		int num, i, sock;
+		return strncpy(buf, "localhost", len);
+#else
+		return strncpy(buf, "127.0.0.1", len);
+#endif
+	}
 
-		my_ifconf.ifc_len = sizeof(my_ifreqs);
-		my_ifconf.ifc_req = my_ifreqs;
-		sock = socket(AF_INET,SOCK_DGRAM,0);
-		if (-1 == sock) 
-			goto out;
+	if ((LINK_NET_ID_IS_IPADDR == which)
+	    || (LINK_NET_ID_IS_CUSTOM == which)) {
+#ifndef G_OS_WIN32
+		if (fixed_host_net_id) 
+			strncpy(buf, fixed_host_net_id, len);
+		else
+			get_first_non_local_ipaddr(buf, len);
+		if ('\0' == *buf)
+			strncpy(buf, "127.0.0.1", len);
 
-		if (ioctl(sock,SIOCGIFCONF,&my_ifconf) < 0) {
-			close(sock);
-			goto out;
-		}
-		close(sock);
-
-		num = my_ifconf.ifc_len / sizeof(struct ifreq);
-		if (!num) 
-			goto out;
-
- 		for (i = 0; i < num; i++) {
-			adr = (struct sockaddr_in *)&my_ifreqs[i].ifr_ifru.ifru_addr;
-			if (strcmp("127.0.0.1", inet_ntoa(adr->sin_addr)))
-				break;
-		}
-		/* will be 127.0.0.1 anyway, if no other address is defined... */
-		return strncpy(buf, (const char*)inet_ntoa(adr->sin_addr), len);
+		return buf;
 #else
 		SOCKET sock;
 		DWORD nbytes;
@@ -200,8 +232,10 @@ get_netid(LinkNetIdType which,
 		for (i = 0; i < nbytes / sizeof (INTERFACE_INFO); i++) {
 			if ((interfaces[i].iiFlags & IFF_UP) &&
 			    !(interfaces[i].iiFlags & IFF_LOOPBACK) &&
-			    interfaces[i].iiAddress.Address.sa_family == AF_INET)
+			    interfaces[i].iiAddress.Address.sa_family == AF_INET) {
+				d_printf("%s:%s:%d: returning %s\n", __FILE__, __FUNCTION__, __LINE__, inet_ntoa(interfaces[i].iiAddress.AddressIn.sin_addr));
 				return strncpy(buf, inet_ntoa(interfaces[i].iiAddress.AddressIn.sin_addr), len);
+			}
 		}
 
 		/* Next look for a loopback IPv4 address */
@@ -218,6 +252,7 @@ get_netid(LinkNetIdType which,
 	}
 
 	if ((LINK_NET_ID_IS_SHORT_HOSTNAME == which) || (LINK_NET_ID_IS_FQDN == which)) {
+                d_printf("%s:%s:%d:gethostname()\n", __FILE__, __FUNCTION__, __LINE__);
 		if (gethostname(buf, len))
 			goto out;
 #ifndef G_OS_WIN32
@@ -249,6 +284,7 @@ get_netid(LinkNetIdType which,
 		struct hostent *he;
 
 		/* gethostbyname() is MT-safe on Windows, btw */
+                d_printf("%s:%s:%d:gethostbyname(%s)\n", __FILE__, __FUNCTION__, __LINE__, buf);
 		he = gethostbyname(buf);
 
 		if (!he)
@@ -282,6 +318,16 @@ link_use_local_hostname (LinkNetIdType use)
         use_local_host = use;
 }
 
+void 
+link_set_local_hostname (const char *host_id)
+{
+	if (!host_id || !strlen(host_id))
+		return;
+
+	fixed_host_net_id = host_id;
+	use_local_host = LINK_NET_ID_IS_CUSTOM;
+}
+
 /*
  * True if succeeded in mapping, else false.
  */
@@ -379,12 +425,17 @@ link_protocol_is_local_ipv46 (const LinkProtocolInfo *proto,
 		 */
 		if (proto->family == AF_INET) {
 			if (af == AF_INET) {
+				struct in_addr ipv4_def_addr;
+
 				struct sockaddr_in * localaddr;
 				struct sockaddr_in * incomingaddr;
+
+				inet_aton ("127.0.0.1", &ipv4_def_addr);
 				localaddr = (struct sockaddr_in *)result->ai_addr;
 				incomingaddr = (struct sockaddr_in *)saddr;
 				localaddr->sin_port = incomingaddr->sin_port;
-				if (!memcmp (localaddr, incomingaddr, result->ai_addrlen)) {
+				if (ipv4_def_addr.s_addr == incomingaddr->sin_addr.s_addr || 
+					!memcmp (localaddr, incomingaddr, result->ai_addrlen)) {
 #ifdef LOCAL_DEBUG
 					g_warning ("local ipv4 address");
 #endif
@@ -394,12 +445,16 @@ link_protocol_is_local_ipv46 (const LinkProtocolInfo *proto,
 		}
 		else {
 			if (af == AF_INET6) {
+				struct in6_addr ipv6_def_addr;
 				struct sockaddr_in6 * localaddr;
 				struct sockaddr_in6 * incomingaddr;
+
+				inet_pton (AF_INET6, "::1", (void *)&ipv6_def_addr);
 				localaddr = (struct sockaddr_in6 *)result->ai_addr;
 				incomingaddr = (struct sockaddr_in6 *)saddr;
 				localaddr->sin6_port = incomingaddr->sin6_port;
-				if (!memcmp (localaddr, incomingaddr, result->ai_addrlen)) {
+				if (!memcmp (&ipv6_def_addr, &incomingaddr->sin6_addr, sizeof(ipv6_def_addr))
+                                    || !memcmp (localaddr, incomingaddr, result->ai_addrlen)) {
 #ifdef LOCAL_DEBUG
 					g_warning ("local ipv6 address");
 #endif
@@ -415,6 +470,7 @@ link_protocol_is_local_ipv46 (const LinkProtocolInfo *proto,
 #else   /*HAVE_GETADDRINFO*/	
 	if (!local_hostent) {
 		LINK_RESOLV_SET_IPV6;
+                d_printf("%s:%s:%d:gethostbyname(%s)\n", __FILE__, __FUNCTION__, __LINE__, link_get_local_hostname ());
 		local_hostent = gethostbyname (link_get_local_hostname ());
 	}
 
@@ -529,6 +585,7 @@ link_protocol_get_sockaddr_ipv4 (const LinkProtocolInfo *proto,
 			res_init();
 #endif
 		
+                d_printf("%s:%s:%d:gethostbyname(%s)\n", __FILE__, __FUNCTION__, __LINE__, hostname);
 		host = gethostbyname (hostname);
 		if (!host) {
 		  g_free (saddr);
@@ -626,6 +683,7 @@ link_protocol_get_sockaddr_ipv6 (const LinkProtocolInfo *proto,
 #endif
 
 	LINK_RESOLV_SET_IPV6;
+        d_printf("%s:%s:%d:gethostbyname(%s)\n", __FILE__, __FUNCTION__, __LINE__, hostname);
 	host = gethostbyname (hostname);
 	if (!host || host->h_addrtype != AF_INET6) {
 		g_free (saddr);
@@ -806,28 +864,14 @@ link_protocol_get_sockinfo_ipv4 (const LinkProtocolInfo  *proto,
 
 	g_assert (proto && saddr && saddr->sa_family == AF_INET);
 
-	if (sa_in->sin_addr.s_addr != INADDR_ANY) {
+#ifdef G_OS_WIN32
+	hname = inet_ntoa(sa_in->sin_addr);
+#else
+	if (ntohl (sa_in->sin_addr.s_addr) != INADDR_ANY) {
 		host = gethostbyaddr ((char *)&sa_in->sin_addr, 
                                       sizeof (struct in_addr), AF_INET);
-		if (!host)
-			return FALSE;
-		hname = host->h_name;
-	}
-#ifdef G_OS_WIN32
-	/* Make sure looking up that name works */
-	if (strcmp (hname, "localhost") == 0)
-		hname = "127.0.0.1"; /* Don't take any risk */
-	else {
-		char *hname_copy = g_strdup (hname);
-		host = gethostbyname (hname_copy);
-		
-		g_free (hname_copy);
-		if (host == NULL) {
-			/* Nope, use IP address then */
-			hname = inet_ntoa (sa_in->sin_addr);
-		} else {
+		if (host)
 			hname = host->h_name;
-		}
 	}
 #endif
 	return link_protocol_get_sockinfo_ipv46 (hname, sa_in->sin_port, 
@@ -886,16 +930,12 @@ link_protocol_get_sockinfo_ipv6 (const LinkProtocolInfo  *proto,
 	if (memcmp (&sa_in6->sin6_addr, &in6addr_any, sizeof (struct in6_addr))) {
 
 #ifdef HAVE_GETNAMEINFO
-                if (getnameinfo((struct sockaddr *)sa_in6, sizeof(*sa_in6), hbuf, sizeof(hbuf), NULL, 0, NI_NAMEREQD))
-			return FALSE;
-		else
+                if (!getnameinfo((struct sockaddr *)sa_in6, sizeof(*sa_in6), hbuf, sizeof(hbuf), NULL, 0, NI_NAMEREQD))
 			hname = hbuf;
 	}
 #else
 		host = gethostbyaddr ((char *)&sa_in6->sin6_addr,
 				      sizeof (struct in6_addr), AF_INET6);
-		if (!host)
-			return FALSE;
 		if (host)
 			hname = host->h_name;
 	}
@@ -949,6 +989,28 @@ link_protocol_get_sockinfo_unix (const LinkProtocolInfo  *proto,
 
 	return TRUE;
 }
+
+/*
+ * link_protocol_post_create_unix:
+ * @fd: the file descriptor.
+ * @sockaddr: a #sockaddr_un structure describing the socket.
+ *
+ * For a user running a program as root, set the owner of the socket to
+ * the original user.
+ */
+static void
+link_protocol_post_create_unix (int fd, struct sockaddr *saddr)
+{
+#ifndef G_OS_WIN32
+  struct sockaddr_un *saddr_un = (struct sockaddr_un *)saddr;
+
+	if (getuid() == 0) {
+		struct stat stat_buf;
+		if (!stat (link_tmpdir, &stat_buf))
+			chown (saddr_un->sun_path, stat_buf.st_uid, -1);
+	}
+#endif
+}
 #endif /* AF_UNIX */
 
 /*
@@ -1086,7 +1148,8 @@ static LinkProtocolInfo static_link_protocols[] = {
 	NULL, 				/* destroy */
 	link_protocol_get_sockaddr_ipv4,/* get_sockaddr */
 	link_protocol_get_sockinfo_ipv4,/* get_sockinfo */
-	link_protocol_is_local_ipv46    /* is_local */
+	link_protocol_is_local_ipv46,   /* is_local */
+	NULL				/* post_create */
 	},
 #endif
 #if defined(AF_INET6)
@@ -1100,7 +1163,8 @@ static LinkProtocolInfo static_link_protocols[] = {
 	NULL, 				/* destroy */
 	link_protocol_get_sockaddr_ipv6,/* get_sockaddr */
 	link_protocol_get_sockinfo_ipv6,/* get_sockinfo */
-	link_protocol_is_local_ipv46    /* is_local */
+	link_protocol_is_local_ipv46,   /* is_local */
+	NULL				/* post_create */
 	},
 #endif
 #ifdef AF_UNIX
@@ -1114,7 +1178,8 @@ static LinkProtocolInfo static_link_protocols[] = {
 	link_protocol_unix_destroy,  			/* destroy */
 	link_protocol_get_sockaddr_unix, 		/* get_sockaddr */
 	link_protocol_get_sockinfo_unix, 		/* get_sockinfo */
-	link_protocol_unix_is_local                     /* is_local */
+	link_protocol_unix_is_local,                     /* is_local */
+	link_protocol_post_create_unix			/* post_create */
 	},
 #endif
 	{ NULL /* name */ }
